#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#define __USE_GNU
#include <sys/syscall.h>
#include <pthread.h>
#include <sched.h>
#include <tracefs.h>

static void trace(const char *fmt, ...)
{
	va_list ap, ap2;

	va_start(ap, fmt);
	va_copy(ap, ap2);
	tracefs_vprintf(NULL, fmt, ap);
	vprintf(fmt, ap2);
	va_end(ap2);
	va_end(ap);
}

/*
 * Three threads A B and C. A is the highest prio
 * C is the lowest.  C starts out and grabs mutex L
 * then A starts, preempting C and it to grabs mutex L
 * So C starts again to finish up, but this time we also
 * wake up B. If B runs and preempts C, we fail,  since
 * that's unbounded priority inversion!
 */
#define MAIN_PRIO 6
#define A_PRIO 5
#define B_PRIO 4
#define C_PRIO 2
#define SLEEP_SECS 5

/* L_lock, the lock that A and C want */
pthread_mutex_t L_lock;

/* ordering of events */
static pthread_barrier_t start_A;
static pthread_barrier_t start_B;
static pthread_barrier_t start_C;
static bool B_spins_a_long_time = true;
static bool B_ran_a_lot;
static bool PI_has_failed;

#define barrier() asm volatile (" " : : : "memory")
#define gettid() syscall(__NR_gettid)

static void set_prio(int prio)
{
	struct sched_param sp = { .sched_priority = prio };
	sched_setscheduler(gettid(), SCHED_FIFO, &sp);
}

static void *thread_A(void *arg)
{
	set_prio(A_PRIO);

	pthread_barrier_wait(&start_A);
	trace("A started\n");

	trace("A waking up B\n");
	pthread_barrier_wait(&start_B);

	trace("A grabbing lock\n");
	pthread_mutex_lock(&L_lock);
	if (B_ran_a_lot)
		PI_has_failed = true;
	trace("A has lock\n");
	pthread_mutex_unlock(&L_lock);
	trace("A released lock\n");

	trace("A exits\n");
	return NULL;
}

static void *thread_B(void *arg)
{
	set_prio(B_PRIO);

	pthread_barrier_wait(&start_B);

	trace("B started\n");

	while (B_spins_a_long_time)
		barrier();

	B_ran_a_lot = 1;

	trace("B exits\n");
	return NULL;
}

static void *thread_C(void *arg)
{
	unsigned long long i;

	set_prio(C_PRIO);

	pthread_barrier_wait(&start_C);
	trace("C started\n");

	pthread_mutex_lock(&L_lock);
	trace("C has lock\n");

	pthread_barrier_wait(&start_A);

	/* spin a little */
	for (i=0; i < 1000000000; i++)
		barrier();

	if (B_ran_a_lot)
		trace("C ran after B\n");

	trace("C releasing lock\n");
	pthread_mutex_unlock(&L_lock);
	trace("C no longer has lock\n");

	trace("C exits\n");
	return NULL;
}

int main (int argc, char **argv)
{
	pthread_mutexattr_t attr;
	struct tep_handle *tep;
	struct tep_event *sched_waking;
	cpu_set_t cpumask;
	pthread_t A,B,C;
	int secs = SLEEP_SECS;
	char pid[24];

	CPU_ZERO(&cpumask);
	CPU_SET(0, &cpumask);
	sched_setaffinity(0, sizeof(cpumask), &cpumask);

	sprintf(pid, "%ld", gettid());

	tracefs_print_init(NULL);
	tracefs_event_disable(NULL, NULL, NULL);
	tracefs_tracer_clear(NULL);
	tracefs_instance_file_write(NULL, "trace", "");
	tracefs_instance_file_write(NULL, "set_event_pid", pid);
	tracefs_option_enable(NULL, TRACEFS_OPTION_EVENT_FORK);

	tep = tracefs_local_events(NULL);
	sched_waking = tep_find_event_by_name(tep, "sched", "sched_waking");
	tracefs_event_filter_apply(NULL, sched_waking, "comm == \"pi-test\"");
	tracefs_event_enable(NULL, "sched", "sched_waking");
	tracefs_event_enable(NULL, "sched", "sched_switch");
	tracefs_event_enable(NULL, "sched", "sched_pi_setprio");

	pthread_mutexattr_init(&attr);
	pthread_mutex_init(&L_lock, &attr);

	pthread_barrier_init(&start_A, NULL, 2);
	pthread_barrier_init(&start_C, NULL, 2);
	pthread_barrier_init(&start_B, NULL, 2);

	pthread_create(&A, NULL, thread_A, NULL);
	pthread_create(&B, NULL, thread_B, NULL);
	pthread_create(&C, NULL, thread_C, NULL);

	set_prio(MAIN_PRIO);

	tracefs_trace_on(NULL);
	trace("Let er rip!\n");
	pthread_barrier_wait(&start_C);

	sleep(secs);

	trace("Stopping B\n");
	B_spins_a_long_time = false;

	trace("wait for A\n");
	pthread_join(A, NULL);
	trace("wait for B\n");
	pthread_join(B, NULL);
	trace("wait for C\n");
	pthread_join(C, NULL);

	tracefs_trace_off(NULL);
	tracefs_print_close(NULL);

	if (PI_has_failed)
		printf("Priority inheritance failed\n");
	else
		printf("Priority inheritance worked like a charm!\n");

	exit(0);
}
