#define _GNU_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <sched.h>
#include <errno.h>
#include <trace-cmd.h>

enum tasks {
	TASK_A,
	TASK_B,
	TASK_C,
};

enum pi_states {
	PI_START,
	PI_GRABBING_LOCK,
	PI_HAS_LOCK,
	PI_RELEASING_LOCK,
	PI_RELEASED_LOCK,
	PI_FINISH,
};

struct pi_state {
	unsigned int	id;
	unsigned int	tid;
	enum tasks	task;
	enum pi_states	state;
};

#define PI_ID 1991

static unsigned int raw_id;
static unsigned int data_offset;

static void process_state(struct pi_state *pstate)
{
	const char *task;
	const char *state;

	switch (pstate->task) {
	case TASK_A: task = "Task A"; break;
	case TASK_B: task = "Task B"; break;
	case TASK_C: task = "Task C"; break;
	}

	switch (pstate->state) {
	case PI_START: state = "staring"; break;
	case PI_GRABBING_LOCK: state = "grabbing lock"; break;
	case PI_HAS_LOCK: state = "has lock"; break;
	case PI_RELEASING_LOCK: state = "releasing lock"; break;
	case PI_RELEASED_LOCK: state = "released lock"; break;
	case PI_FINISH: state = "finished"; break;
	}

	printf("[%d] %s %s\n", pstate->tid, task, state);
}

static int print_record(struct tep_handle *tep, struct tep_record *record,
			struct trace_seq *seq)
{
	struct pi_state *state;
	int cpu = record->cpu;

	if (tep_data_type(tep, record) == raw_id) {
		if (record->size - data_offset >= sizeof(*state)) {
			state = record->data + data_offset;
			if (state->id == PI_ID)
				process_state(state);
		}
		return 0;
	}
	trace_seq_reset(seq);
	tep_print_event(tep, seq, record, "%6.1000d", TEP_PRINT_TIME);
	trace_seq_printf(seq, " [%03d] ", cpu);
	tep_print_event(tep, seq, record, "%s-%d %s %s\n",
			TEP_PRINT_COMM, TEP_PRINT_PID,
			TEP_PRINT_NAME, TEP_PRINT_INFO);
	trace_seq_do_printf(seq);
	return 0;
}

static struct tep_record *get_next_record(struct tep_handle *tep,
					  struct tracecmd_input *handle)
{
	static struct tep_record **cpu_records;
	struct tep_record *record;
	unsigned long long ts = -1ULL;
	int nr_cpus;
	int next = -1;
	int i;

	nr_cpus = tep_get_cpus(tep);

	if (!cpu_records) {
		cpu_records = calloc(nr_cpus, sizeof(*cpu_records));
		for (i = 0; i < nr_cpus; i++)
			cpu_records[i] = tracecmd_read_cpu_first(handle, i);
	}

	for (i = 0; i < nr_cpus; i++) {
		if (cpu_records[i] && cpu_records[i]->ts < ts)
			next = i;
	}
	if (next < 0)
		return NULL;

	record = cpu_records[next];
	cpu_records[next] = tracecmd_read_data(handle, next);

	return record;
}

int main(int argc, char **argv) {
	struct tracecmd_input *handle, *pi_handle;
	struct tep_record *record;
	struct tep_event *raw_event;
	struct tep_format_field *id;
	struct tep_handle *tep;
	struct trace_seq seq;
	char *file = "trace.dat";
	int i, nr_instances;

	if (argc > 1)
		file = argv[1];

	tracecmd_set_loglevel(TEP_LOG_CRITICAL);

	handle = tracecmd_open(file, 0);
	nr_instances = tracecmd_buffer_instances(handle);
	for (i = 0; i < nr_instances; i++) {
		if (strcmp("pi-test", tracecmd_buffer_instance_name(handle, i)) == 0)
			break;
	}
	pi_handle = tracecmd_buffer_instance_handle(handle, i);

	tep = tracecmd_get_tep(pi_handle);
	trace_seq_init(&seq);

	raw_event = tep_find_event_by_name(tep, "ftrace", "raw_data");
	raw_id = raw_event->id;
	id = tep_find_field(raw_event, "id");
	data_offset = id->offset;

	while ((record = get_next_record(tep, pi_handle)) != NULL) {
		print_record(tep, record, &seq);
		tracecmd_free_record(record);
	}
	trace_seq_destroy(&seq);

	return 0;
}
